"""transaction.py - High-level access to PayPal transactions"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum

from . import errors

from decimal import Decimal

from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
)
from .paypal_types import (
    Amount,
    APIResponse,
    parse_datetime,
)

T = TypeVar('T')

class CartItem(NamedTuple):
    code: Optional[str]
    name: Optional[str]
    description: Optional[str]
    quantity: Union[int, Decimal]
    unit_price: Amount
    total_price: Amount

    @classmethod
    def from_api(cls, source: APIResponse, default_name: Optional[str]=None) -> 'CartItem':
        total_price = Amount.from_api(source['item_amount'])
        quantity = Decimal(source.get('item_quantity', 1))
        try:
            unit_price = Amount.from_api(source['item_unit_price'])
        except KeyError:
            unit_price = total_price._replace(number=total_price.number / quantity)
        return cls(
            source.get('item_code'),
            source.get('item_name', default_name),
            source.get('item_description'),
            quantity,
            unit_price,
            total_price,
        )


class TransactionStatus(enum.Enum):
    D = "Denied"
    DENIED = D
    F = "Partially Refunded"
    REFUNDED = F
    P = "Pending"
    PENDING = P
    S = "Successful"
    SUCCESSFUL = S
    SUCCESS = S
    V = "Reversed"
    REVERSED = V


class Transaction(APIResponse):
    """PayPal Transaction wrapper

    The public methods of a Transaction know how to traverse PayPal's JSON
    response and turn the data into native Python objects. If Transaction is
    missing a method you need, you can also use standarding Mapping methods to
    access to raw response.
    """
    __slots__ = ['_response']

    def __init__(self, response: APIResponse) -> None:
        self._response = response

    def __getitem__(self, key: str) -> Any:
        return self._response[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._response)

    def __len__(self) -> int:
        return len(self._response)

    def _get_from_response(self, *keys: str) -> Any:
        retval = self._response
        for index, key in enumerate(keys):
            try:
                retval = retval[key]
            except KeyError as error:
                try:
                    txn_id = f"Transaction {self['transaction_info']['transaction_id']}"
                except KeyError:
                    txn_id = "Transaction"
                if index:
                    key_s = '→'.join(repr(key) for key in keys[:index + 1])
                    raise KeyError(f"{txn_id} {key_s}") from None
                else:
                    raise errors.MissingFieldError(
                        f"{txn_id} was not loaded with {error.args[0]!r} field",
                    ) from None
        return retval

    def _wrap_response(  # type:ignore[misc]
            name: str,
            func: Callable[[Any], T],
            *keys: str,
            doc: Optional[str]=None,
            key_doc: Optional[str]=None,
            return_doc: Optional[str]=None,
    ) -> Callable[['Transaction'], T]:
        if doc is None:
            if key_doc is None:
                key_doc = f"``{keys[-1]}``"
            if return_doc is None:
                if name.endswith('amount'):
                    return_doc = 'Amount'
                elif name.endswith('date'):
                    return_doc = 'datetime'
                else:
                    return_doc = func.__name__
            doc = f"Return the transaction's {key_doc} as a ``{return_doc}``"
        def response_wrapper(self: 'Transaction') -> T:
            return func(self._get_from_response(*keys))
        response_wrapper.__name__ = name
        response_wrapper.__doc__ = doc
        return response_wrapper

    def _fee_amount(txn_info: APIResponse) -> Optional[Amount]:  # type:ignore[misc]
        try:
            raw_fee = txn_info['fee_amount']
        except KeyError:
            return None
        else:
            return Amount.from_api(raw_fee)

    amount = _wrap_response(
        'amount',
        Amount.from_api,
        'transaction_info',
        'transaction_amount',
    )
    fee_amount = _wrap_response(
        'fee_amount',
        _fee_amount,
        'transaction_info',
        key_doc='``fee_amount``',
    )
    initiation_date = _wrap_response(
        'initiation_date',
        parse_datetime,
        'transaction_info',
        'transaction_initiation_date',
    )
    payer_email = _wrap_response(
        'payer_email',
        str,
        'payer_info',
        'email_address',
        key_doc="payer's email address",
    )
    payer_fullname = _wrap_response(
        'payer_fullname',
        str,
        'payer_info',
        'payer_name',
        'alternate_full_name',
        key_doc="payer's full name",
    )
    status = _wrap_response(
        'status',
        TransactionStatus.__getitem__,
        'transaction_info',
        'transaction_status',
        return_doc=TransactionStatus.__name__,
    )
    transaction_id = _wrap_response(
        'transaction_id',
        str,
        'transaction_info',
        'transaction_id',
    )
    updated_date = _wrap_response(
        'updated_date',
        parse_datetime,
        'transaction_info',
        'transaction_updated_date',
    )

    def cart_items(self) -> Iterator[CartItem]:
        """Iterate a ``CartItem`` object for each item in the transaction's cart"""
        cart_info = self._get_from_response('cart_info')
        try:
            item_seq = cart_info['item_details']
        except KeyError:
            pass
        else:
            try:
                default_name = self['transaction_info']['transaction_subject']
            except KeyError:
                default_name = None
            for source in item_seq:
                try:
                    yield CartItem.from_api(source, default_name)
                except KeyError:
                    pass



"""__init__.py - paypal_rest unit test common functionality"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import typing
import urllib.parse as urlparse

import requests

FLOAT_NUMBERS = [
    '1200.30',
    '255.75',
    '99.99',
    '5.00',
    '0.00',
]
FLOAT_NUMBERS.extend(f'-{s}' for s in FLOAT_NUMBERS[:-1])

WHOLE_NUMBERS = [
    '1250',
    '255',
    '99',
    '5',
    '0',
]
WHOLE_NUMBERS.extend(f'-{s}' for s in WHOLE_NUMBERS[:-1])

class ReceivedRequest(typing.NamedTuple):
    method: str
    url: urlparse.ParseResult
    params: typing.Mapping[str, str]

    @classmethod
    def from_args(cls, method, url, params):
        return cls(
            method,
            urlparse.urlparse(url),
            dict(params or ()),
        )


class MockResponse:
    def __init__(self, body, status_code=requests.codes.OK):
        self._body = body
        self.status_code = status_code

    @classmethod
    def error(
            cls,
            name='TEST_ERROR',
            message='Test error',
            debug_id='0000',
            status_code=requests.codes.BAD_REQUEST,
            *details,
            **kwargs,
    ):
        body = {
            'name': name,
            'message': message,
            'debug_id': debug_id,
            'details': [{'issue': issue, 'location': 'test'}
                        for issue in details],
        }
        body.update(kwargs)
        return cls(body, status_code)

    def json(self):
        return self._body.copy()

    def raise_for_status(self):
        if self.status_code >= 400:
            raise requests.HTTPError(self.status_code)


class MockSession:
    LAST_RESPONSE = MockResponse({
        'name': 'NO_RESPONSE',
        'message': 'MockSession got more requests than expected',
    }, 509)

    def __init__(self, *responses, infinite=False):
        self._requests = []
        if infinite:
            self._responses = itertools.cycle(responses)
        else:
            self._responses = itertools.chain(
                responses,
                itertools.repeat(self.LAST_RESPONSE),
            )

    def request(self, method, url, params=None):
        self._requests.append(ReceivedRequest.from_args(method, url, params))
        return next(self._responses)

    def get(self, url, params=None):
        return self.request('GET', url, params)



"""client.py - PayPal API client class"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import datetime
import enum
import logging
import math
import operator
import urllib.parse as urlparse

import requests
import requests_oauthlib  # type:ignore[import]
from oauthlib import oauth2  # type:ignore[import]

from typing import (
    Any,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)
from .paypal_types import (
    APIResponse,
)
from .transaction import Transaction

FieldsType = TypeVar('FieldsType', bound='PayPalFields')
Params = Mapping[str, str]

class PayPalSite(enum.Enum):
    SANDBOX = 'api-m.sandbox.paypal.com'
    LIVE = 'api-m.paypal.com'

    def url(self) -> str:
        return f'https://{self.value}'


class PayPalFields(enum.Flag):
    """Base class for PayPal fields specifiers

    Multiple PayPal APIs accept a ``fields`` parameter to let the user specify
    what details to return. This class lets code enumerate acceptable values
    and combine them programmatically. The ``param_value`` method then helps
    ``PayPalAPIClient`` format the result when needed.
    """

    @classmethod
    def choices(cls) -> Iterator[str]:
        """Iterate the names of all field values"""
        for flag in cls:
            yield flag.name.lower()

    @classmethod
    def combine(cls: Type[FieldsType], fields: Optional[Sequence[FieldsType]]=None) -> FieldsType:
        """Combine multiple field objects into one

        This method just returns the result of ORing all of the fields in the
        sequence together. If no argument is given or the sequence is empty,
        return the combination of all fields.
        """
        if fields:
            fields_iter = iter(fields)
        else:
            fields_iter = iter(cls)
        retval = next(fields_iter)
        for field in fields_iter:
            retval |= field
        return retval

    @classmethod
    def from_arg(cls: Type[FieldsType], arg: str) -> FieldsType:
        """Return a field object from an argument name string"""
        try:
            return cls[arg.upper()]
        except KeyError:
            raise ValueError(f"unknown {cls.__name__} {arg!r}") from None

    def is_base_field(self) -> bool:
        """Return true if this is a single field value, not a combination"""
        return not math.log2(self.value) % 1

    def _base_value(self) -> str:
        return self.name.lower()

    def param_value(self) -> str:
        """Return these fields formatted as a query string

        The result is in the format PayPal's API expects for ``fields``
        parameters.
        """
        return ','.join(
            flag._base_value()
            for flag in type(self)
            if flag & self and flag.is_base_field()
        )


class SubscriptionFields(PayPalFields):
    LAST_FAILED_PAYMENT = enum.auto()
    PLAN = enum.auto()
    ALL = LAST_FAILED_PAYMENT | PLAN


class TransactionFields(PayPalFields):
    TRANSACTION = enum.auto()
    PAYER = enum.auto()
    SHIPPING = enum.auto()
    AUCTION = enum.auto()
    CART = enum.auto()
    INCENTIVE = enum.auto()
    STORE = enum.auto()
    ALL = TRANSACTION | PAYER | SHIPPING | AUCTION | CART | INCENTIVE | STORE

    def _base_value(self) -> str:
        return f'{self.name.lower()}_info'


class PayPalSession(requests_oauthlib.OAuth2Session):
    """Low-level HTTP session for the PayPal API

    This is a subclass of requests_oauthlib.OAuth2Session that implements
    PayPal's recommended authorization strategy: if an API request returns
    HTTP Unauthorized, get an OAuth token and retry. This gracefully handles
    refreshing expired tokens.

    This class only handles the mechanics of handling an HTTP connection.
    It doesn't know anything about the higher-level REST API. That's the job
    of ``PayPalAPIClient``.
    """
    TOKEN_PATH = '/v1/oauth2/token'

    def __init__(self, client: oauth2.Client, client_secret: str) -> None:
        super().__init__(client=client)
        self._client_secret = client_secret

    def request(self, method: str, url: str, **kwargs: Any) -> requests.Response:
        response = super().request(method, url, **kwargs)
        if response.status_code == requests.codes.UNAUTHORIZED:
            self.fetch_token(
                token_url=urlparse.urljoin(url, self.TOKEN_PATH),
                client_id=self._client.client_id,
                client_secret=self._client_secret,
            )
            response = super().request(method, url, **kwargs)
        return response


class PayPalAPIClient:
    """Primary access point for the PayPal API

    This is the primary class of the library. Most users will instantiate a
    ``PayPalAPIClient`` using one of the constructor classmethods, then call
    methods to make API calls.
    """

    def __init__(
            self,
            session: requests.Session,
            root_url: Union[str, PayPalSite]=PayPalSite.SANDBOX,
            logger: Optional[logging.Logger]=None,
    ) -> None:
        """Low-level constructor

        ``PayPalAPIClient`` expects its underlying ``Session`` object to know
        how to authorize itself to PayPal. Usually that means using an instance
        of ``PayPalSession``. You can implement and provide your own subclass
        of ``requests.Session`` to do this if you prefer.

        ``root_url`` is either a PayPalSite value, or a string with a full URL
        to a PayPal API endpoint.

        ``logger`` is a ``logging.Logger`` object where all log messages (like
        API errors) will be sent. If none is provided, this instance will get
        its own, with a name based on the hostname in ``root_url``.
        """
        self._session = session
        if isinstance(root_url, str):
            self._root_url = root_url
        else:
            self._root_url = root_url.url()
        if logger is None:
            host_parts = urlparse.urlsplit(self._root_url).netloc.split('.')
            host_parts.reverse()
            logger_name = f'paypal_rest.PayPalAPIClient.{".".join(host_parts)}'
            logger = logging.getLogger(logger_name)
        self.logger = logger

    @classmethod
    def from_client_secret(
            cls,
            client_id: str,
            client_secret: str,
            root_url: Union[str, PayPalSite]=PayPalSite.SANDBOX,
            logger: Optional[logging.Logger]=None,
    ) -> 'PayPalAPIClient':
        """High-level constructor from individual string arguments

        Given ``client_id`` and ``client_secret`` strings, this method
        constructs a ``PayPalSesssion`` from them, and then returns a
        ``PayPalAPIClient`` backed by it.

        ``root_url`` and ``logger`` arguments are passed directly to
        ``PayPalAPIClient.__init__``.
        """
        client = oauth2.BackendApplicationClient(client_id=client_id)
        session = PayPalSession(client, client_secret)
        return cls(session, root_url, logger)

    @classmethod
    def from_config(
            cls,
            config: Mapping[str, str],
            default_url: Union[str, PayPalSite]=PayPalSite.SANDBOX,
            logger: Optional[logging.Logger]=None,
    ) -> 'PayPalAPIClient':
        """High-level constructor from a configuration mapping

        Given a mapping of strings (e.g., a configparser section object),
        gets the arguments necessary to call ``from_client_secret``, and calls
        it.

        If the mapping has a ``site`` key, the value will be used as the
        ``root_url``. Otherwise, ``root_url`` has the value of ``default_url``.

        ``logger`` is passed directly to ``PayPalAPIClient.__init__``.
        """
        try:
            client_id = config['client_id']
            client_secret = config['client_secret']
        except KeyError as error:
            raise ValueError(f"configuration missing {error.args[0]!r}") from None
        try:
            root_url: Union[str, PayPalSite] = config['site']
        except KeyError:
            root_url = default_url
        else:
            try:
                # In this case, we know root_url must be a str because we got
                # it from the Mapping, but mypy doesn't track that.
                root_url = PayPalSite[root_url.upper()]  # type:ignore[union-attr]
            except KeyError:
                pass
        return cls.from_client_secret(client_id, client_secret, root_url, logger)

    def _get_json(self, path: str, params: Optional[Params]=None) -> APIResponse:
        url = urlparse.urljoin(self._root_url, path)
        response = self._session.get(url, params=params)
        try:
            response.raise_for_status()
        except requests.HTTPError:
            self._log_error(response.json())
            raise
        else:
            return response.json()

    def _iter_pages(self, path: str, params: Params) -> Iterator[APIResponse]:
        response: APIResponse = {'page': 0, 'total_pages': 1}
        page_params = collections.ChainMap(params).new_child()
        while response['page'] < response['total_pages']:
            page_params['page'] = str(response['page'] + 1)
            response = self._get_json(path, page_params)
            yield response

    def _log_error(self, error: APIResponse) -> None:
        parts = [
            '{name}: {message}'.format_map(error),
            *('{issue} (in {location})'.format_map(detail)
              for detail in error.get('details', ())),
        ]
        self.logger.error(" — ".join(parts))

    def _iter_date_params(
            self,
            start_date: datetime.datetime,
            end_date: datetime.datetime,
            params: Optional[Params]=None,
    ) -> Iterator[Params]:
        """Generate parameters with date windows to cover a wider span

        The PayPal transaction search API only allows the ``start_date`` and
        ``end_date`` to be a month apart. Given a user's desired date range and
        other API parameters, this method generates parameters with different
        date pairs to cover the entire date range.

        Normally the method keeps incrementing ``start_date`` until the user's
        desired ``end_date`` is reached. If the ``start_date`` if later than the
        ``end_date``, instead it works backwards: it uses the ``start_date``
        argument as the first ``end_date`` parameter, and keeps decrementing it
        until the ``start_date`` parameter reaches the other end of the range.
        """
        if start_date > end_date:
            key1 = 'end_date'
            key2 = 'start_date'
            days_sign = operator.neg
            pred = operator.gt
            limit_func = max
        else:
            key1 = 'start_date'
            key2 = 'end_date'
            days_sign = operator.pos
            pred = operator.lt
            limit_func = min
        retval = collections.ChainMap(params or {}).new_child()
        retval[key1] = start_date.isoformat(timespec='seconds')
        next_date = start_date
        date_diff = datetime.timedelta(days=days_sign(30))
        while pred(next_date, end_date):
            next_date = limit_func(next_date + date_diff, end_date)
            retval[key2] = next_date.isoformat(timespec='seconds')
            yield retval
            retval[key1] = retval[key2]

    def get_subscription(
            self,
            subscription_id: str,
            fields: SubscriptionFields=SubscriptionFields.ALL,
    ) -> APIResponse:
        """Fetch and return a subscription by its id"""
        return self._get_json(f'/v1/billing/subscriptions/{subscription_id}', {
            'fields': fields.param_value(),
        })

    def get_transaction(
            self,
            transaction_id: str,
            end_date: Optional[datetime.datetime]=None,
            start_date: Optional[datetime.datetime]=None,
            fields: TransactionFields=TransactionFields.TRANSACTION,
    ) -> Transaction:
        """Find and return a transaction by its id

        The PayPal API does not provide a way to look up transactions solely by
        id. This is a convenience method that wraps the search method to search
        different windows of time until it finds the desired transaction.

        ``start_date`` and ``end_date`` specify the full window of time to
        search. This method starts by searching 30 days before ``end_date``,
        then the previous 30 days, and so on until it reaches ``start_date``.
        The default ``end_date`` is now, and the default ``start_date`` is
        three years ago (the API only supports searches this far back).

        ``fields`` is a TransactionFields object that flags the information to
        include in the returned Transaction.
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        if end_date is None:
            end_date = now
        if start_date is None:
            # The API only goes back three years
            start_date = now - datetime.timedelta(days=365 * 3)
        response: APIResponse = {'transaction_details': None}
        for params in self._iter_date_params(end_date, start_date, {
                'transaction_id': transaction_id,
                'fields': fields.param_value(),
        }):
            response = self._get_json('/v1/reporting/transactions', params)
            if response['transaction_details']:
                return Transaction(response['transaction_details'][0])
        raise ValueError(f"transaction {transaction_id!r} not found")

    def iter_transactions(
            self,
            start_date: datetime.datetime,
            end_date: datetime.datetime,
            fields: TransactionFields=TransactionFields.TRANSACTION,
    ) -> Iterator[Transaction]:
        """Iterate transactions over a date range

        ``start_date`` and ``end_date`` represent the range to query.
        ``fields`` is a TransactionsFields object that flags the information to
        include in returned Transactions.
        """
        for params in self._iter_date_params(start_date, end_date, {
                'fields': fields.param_value(),
        }):
            for page in self._iter_pages('/v1/reporting/transactions', params):
                for txn_source in page['transaction_details']:
                    yield Transaction(txn_source)



"""test_transaction.py - Unit tests for high-level Transaction access"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import re

from decimal import Decimal

import pytest
import requests

from . import MockSession, MockResponse, FLOAT_NUMBERS, WHOLE_NUMBERS

from paypal_rest import errors
from paypal_rest import transaction as txn_mod

DATE1 = datetime.datetime(2020, 10, 2, 14, 15, 16, tzinfo=datetime.timezone.utc)
DATE2 = DATE1.replace(second=56)

def item_source(
        name=None,
        description=None,
        code=None,
        quantity=1,
        number='15.99',
        currency='USD',
):
    if description is None:
        description = name
    if code is None:
        code = name
    total_number = str(quantity * Decimal(number))
    retval = {
        'item_quantity': str(quantity),
        'item_unit_price': {'value': number, 'currency_code': currency},
        'item_item_amount': {'value': total_number, 'currency_code': currency},
        'item_amount': {'value': total_number, 'currency_code': currency},
    }
    if name is not None:
        retval['item_name'] = name
    if description is not None:
        retval['item_description'] = description
    if code is not None:
        retval['item_code'] = code
    return retval

def cart_info(*item_kwargs):
    return {'item_details': [item_source(**kwargs) for kwargs in item_kwargs]}

def payer_info(
        given_name='Payer',
        surname='Smith',
        account_id='PAYER12345678',
        payer_status='Y',
        address_status='Y',
        country_code='US',
        email_address='payer@example.org',
):
    retval = locals().copy()
    del retval['given_name'], retval['surname']
    retval['payer_name'] = {
        'given_name': given_name,
        'surname': surname,
        'alternate_full_name': f'{given_name} {surname}',
    }
    return retval

def transaction_info(
        transaction_id='TRANSACTION123456',
        number='5.00',
        currency='USD',
        fee_number='-0.49',
        fee_currency=None,
        init_date=DATE1,
        update_date=None,
        status='S',
        subject=None,
        note=None,
):
    retval = {
        'transaction_id': transaction_id,
        'transaction_status': status,
        'transaction_initiation_date': init_date.isoformat(timespec='seconds'),
        'transaction_updated_date': (update_date or init_date).isoformat(timespec='seconds'),
        'transaction_amount': {'value': number, 'currency_code': currency},
        'fee_amount': {'value': fee_number, 'currency_code': fee_currency or currency},
    }
    if subject is not None:
        retval['transaction_subject'] = subject
    if note is not None:
        retval['transaction_note'] = note
    return retval

def test_mapping():
    source = {
        'transaction_info': transaction_info(),
        'payer_info': payer_info(),
    }
    txn = txn_mod.Transaction(source.copy())
    assert len(txn) == len(source)
    assert frozenset(txn) == frozenset(source)
    assert txn == source
    assert txn.get('test_key') is None

@pytest.mark.parametrize('number', FLOAT_NUMBERS)
def test_amount_float(number):
    source = transaction_info(number=number, currency='USD')
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.amount() == (Decimal(number), 'USD')

@pytest.mark.parametrize('number', WHOLE_NUMBERS)
def test_amount_whole(number):
    source = transaction_info(number=number, currency='JPY')
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.amount() == (int(number), 'JPY')

def test_cart_items():
    source = cart_info(
        {'number': '5.99'},
        {'name': 'Test', 'number': '7.99', 'quantity': 2},
    )
    txn = txn_mod.Transaction({'cart_info': source})
    item1, item2 = txn.cart_items()
    item1_price = (Decimal('5.99'), 'USD')
    assert item1 == (
        None, None, None, 1, item1_price, item1_price,
    )
    assert item2 == (
        'Test', 'Test', 'Test', 2,
        (Decimal('7.99'), 'USD'),
        (Decimal('15.98'), 'USD'),
    )

@pytest.mark.parametrize('subject', [
    'test subject',
    '$6.99 Subscription',
    None,
])
def test_unnamed_cart_item_uses_transaction_subject(subject):
    source = {
        'cart_info': cart_info({'number': '6.99'}),
        'transaction_info': transaction_info(subject=subject),
    }
    txn = txn_mod.Transaction(source)
    item1, = txn.cart_items()
    assert item1.name == subject

def test_cart_empty():
    txn = txn_mod.Transaction({'cart_info': {}})
    assert not any(txn.cart_items())

def test_cart_quantity_only():
    # Refunds are structured this way.
    source = {'item_details': [{'item_quantity': '1.000'}]}
    txn = txn_mod.Transaction({'cart_info': source})
    assert not any(txn.cart_items())

@pytest.mark.parametrize('number', FLOAT_NUMBERS)
def test_fee_amount_float(number):
    source = transaction_info(fee_number=number, fee_currency='USD')
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.fee_amount() == (Decimal(number), 'USD')

@pytest.mark.parametrize('number', WHOLE_NUMBERS)
def test_fee_amount_whole(number):
    source = transaction_info(fee_number=number, fee_currency='JPY')
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.fee_amount() == (int(number), 'JPY')

def test_fee_amount_none():
    txn = txn_mod.Transaction({'transaction_info': {}})
    assert txn.fee_amount() is None

def test_payer_email():
    email = 'test@example.net'
    source = payer_info(email_address=email)
    txn = txn_mod.Transaction({'payer_info': source})
    assert txn.payer_email() == email

@pytest.mark.parametrize('given_name', ['Robin', 'Sawyer'])
def test_payer_fullname(given_name):
    txn = txn_mod.Transaction({'payer_info': payer_info(given_name)})
    assert txn.payer_fullname() == f'{given_name} Smith'

def test_initiation_date():
    source = transaction_info(init_date=DATE2)
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.initiation_date() == DATE2

@pytest.mark.parametrize('key', 'DFPSV')
def test_status(key):
    source = transaction_info(status=key)
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.status() is txn_mod.TransactionStatus[key]

def test_transaction_id():
    test_id = 'TESTTXNID12345678'
    source = transaction_info(transaction_id=test_id)
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.transaction_id() == test_id

def test_updated_date():
    source = transaction_info(update_date=DATE2)
    txn = txn_mod.Transaction({'transaction_info': source})
    assert txn.updated_date() == DATE2

@pytest.mark.parametrize('method_name', [
    'amount',
    'cart_items',
    'fee_amount',
    'payer_email',
    'payer_fullname',
    'initiation_date',
    'status',
    'transaction_id',
    'updated_date',
])
def test_info_missing(method_name):
    txn = txn_mod.Transaction({})
    with pytest.raises(errors.MissingFieldError):
        result = getattr(txn, method_name)()
        try:
            do_next = iter(result) is result
        except TypeError:
            do_next = False
        if do_next:
            next(result)



"""test_client.py - Unit tests for PayPal API client class"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import re

import pytest
import requests

from . import MockSession, MockResponse

from paypal_rest import client as client_mod
from paypal_rest import transaction as txn_mod

START_DATE = datetime.datetime(2020, 10, 1, 12, tzinfo=datetime.timezone.utc)
END_DATE = START_DATE.replace(day=25)

def test_transaction_type():
    txn_id = 'TYPETEST123456789'
    session = MockSession(
        MockResponse({'page': 1, 'total_pages': 1, 'transaction_details': [
            {'transaction_info': {'transaction_id': txn_id}},
        ]}),
    )
    paypal = client_mod.PayPalAPIClient(session)
    txn, = paypal.iter_transactions(START_DATE, END_DATE)
    assert isinstance(txn, txn_mod.Transaction)
    assert txn.transaction_id() == txn_id

@pytest.mark.parametrize('pages', range(1, 4))
def test_transaction_pagination(pages):
    session = MockSession(
        *(MockResponse({'page': page, 'total_pages': pages, 'transaction_details': []})
          for page in range(1, pages + 1))
    )
    paypal = client_mod.PayPalAPIClient(session)
    actual = list(paypal.iter_transactions(START_DATE, END_DATE))
    assert len(session._requests) == pages
    for expect_page, request in enumerate(session._requests, 1):
        assert request.params.get('page') == str(expect_page)

@pytest.mark.parametrize('index,key', enumerate(['start_date', 'end_date']))
def test_transaction_date_formatting(index, key):
    session = MockSession(
        MockResponse({'page': 1, 'total_pages': 1, 'transaction_details': []}),
    )
    paypal = client_mod.PayPalAPIClient(session)
    args = [START_DATE, END_DATE]
    new_date = START_DATE.replace(day=15)
    args[index] = new_date
    actual = list(paypal.iter_transactions(*args))
    assert session._requests
    assert session._requests[0].params.get(key) == new_date.isoformat(timespec='seconds')

@pytest.mark.parametrize('fields', [
    {'transaction'},
    {'cart'},
    {'transaction', 'payer'},
    {'cart', 'transaction', 'payer'},
])
def test_transaction_fields_formatting(fields):
    fields_iter = (name.upper() for name in fields)
    fields_arg = client_mod.TransactionFields[next(fields_iter)]
    for name in fields_iter:
        fields_arg |= client_mod.TransactionFields[name]
    session = MockSession(
        MockResponse({'page': 1, 'total_pages': 1, 'transaction_details': []}),
    )
    paypal = client_mod.PayPalAPIClient(session)
    _ = list(paypal.iter_transactions(START_DATE, END_DATE, fields_arg))
    assert session._requests
    assert fields == {
        re.sub(r'_info$', '', name)
        for name in session._requests[0].params.get('fields', '').split(',')
    }

@pytest.mark.parametrize('days_diff', [10, 45, 89])
def test_iter_transactions_date_window(days_diff):
    start_date = START_DATE
    end_date = start_date + datetime.timedelta(days=days_diff)
    session = MockSession(
        MockResponse({'page': 1, 'total_pages': 1, 'transaction_details': []}),
        infinite=True,
    )
    paypal = client_mod.PayPalAPIClient(session)
    assert not any(paypal.iter_transactions(start_date, end_date))
    req_count = len(session._requests)
    assert req_count == ((days_diff // 30) + 1)
    start_str = start_date.isoformat(timespec='seconds')
    end_str = end_date.isoformat(timespec='seconds')
    prev_end = start_str
    for number, request in enumerate(session._requests, 1):
        assert request.params['start_date'] == prev_end
        if number == req_count:
            assert request.params['end_date'] == end_str
        else:
            assert prev_end < request.params['end_date'] < end_str
            prev_end = request.params['end_date']

@pytest.mark.parametrize('days_diff', [10, 45, 89])
def test_get_transactions_date_window(days_diff):
    end_date = END_DATE
    start_date = end_date - datetime.timedelta(days=days_diff)
    session = MockSession(
        MockResponse({'page': 1, 'total_pages': 1, 'transaction_details': []}),
        infinite=True,
    )
    paypal = client_mod.PayPalAPIClient(session)
    with pytest.raises(ValueError):
        paypal.get_transaction('DATEWINDOW1234567', end_date, start_date)
    req_count = len(session._requests)
    assert req_count == ((days_diff // 30) + 1)
    start_str = start_date.isoformat(timespec='seconds')
    end_str = end_date.isoformat(timespec='seconds')
    prev_start = end_str
    for number, request in enumerate(session._requests, 1):
        assert request.params['end_date'] == prev_start
        if number == req_count:
            assert request.params['start_date'] == start_str
        else:
            assert prev_start > request.params['start_date'] > start_str
            prev_start = request.params['start_date']

@pytest.mark.parametrize('name', [
    'BAD_REQUEST',
    'UNAUTHORIZED',
    'FORBIDDEN',
    'NOT_FOUND',
    'INTERNAL_SERVER_ERROR',
])
def test_error_handling(name, caplog):
    session = MockSession(MockResponse.error(name, status_code=requests.codes[name]))
    paypal = client_mod.PayPalAPIClient(session)
    with pytest.raises(requests.HTTPError):
        paypal.get_transaction('ABCDEFGHIJKLMNOPQ')
    assert any(
        log.levelname == 'ERROR'
        and log.message == f'{name}: Test error'
        for log in caplog.records
    )



"""errors.py - PayPal API client exception classes"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

class MissingFieldError(KeyError):
    """Error raised when code tries to access an unloaded field

    This error is raised by PayPal object classes when the caller tries to
    access a field that was not loaded in the original API call. For
    example, trying to get a payer's name or email address from a Transaction
    when the ``fields`` argument did not include ``TransactionFields.PAYER``.
    """
    pass



"""test_paypal_types.py - Unit tests for common PayPal API classes"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools

from decimal import Decimal

import pytest
from . import FLOAT_NUMBERS, WHOLE_NUMBERS

from paypal_rest import paypal_types

@pytest.mark.parametrize('number,currency', itertools.product(
    FLOAT_NUMBERS, ['USD', 'INR'],
))
def test_float_amount_from_api(number, currency):
    source = {'value': number, 'currency_code': currency}
    actual = paypal_types.Amount.from_api(source)
    assert actual.number == Decimal(number)
    assert actual.currency == currency

@pytest.mark.parametrize('number,currency', itertools.product(
    WHOLE_NUMBERS, ['JPY', 'CLP'],
))
def test_whole_amount_from_api(number, currency):
    source = {'value': number, 'currency_code': currency}
    actual = paypal_types.Amount.from_api(source)
    assert actual.number == int(number)
    assert actual.currency == currency



"""config.py - Load user configuration for PayPal API client"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import configparser

from pathlib import Path
from xdg import BaseDirectory  # type:ignore[import]

from typing import (
    Union,
)

CONFIG_PATH = Path('paypal_rest', 'config.ini')

def load_config(path: Union[Path, str, None]=None) -> configparser.ConfigParser:
    if path is None:
        path = BaseDirectory.load_first_config(str(CONFIG_PATH))
        if path is None:
            # This path doesn't exist, but we'll handle that case later.
            path = Path(BaseDirectory.xdg_config_home, CONFIG_PATH)
    config = configparser.ConfigParser()
    with open(path) as config_file:
        config.read_file(config_file)
    return config



"""paypal_types.py - Common types and classes across the PayPal API"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime

from decimal import Decimal

from typing import (
    Any,
    Callable,
    Mapping,
    NamedTuple,
)

# Declare a type definition for the one function of iso8601 that we use.
import iso8601  # type:ignore[import]
parse_datetime: Callable[[str], datetime.datetime] = iso8601.parse_date

APIResponse = Mapping[str, Any]

class Amount(NamedTuple):
    number: Decimal
    currency: str

    @classmethod
    def from_api(cls, source: APIResponse) -> 'Amount':
        return cls(
            Decimal(source['value']),
            source['currency_code'],
        )

    def __str__(self) -> str:
        return '{:,g} {}'.format(*self)



"""cliutil - Utilities for CLI tools"""
PKGNAME = 'paypal_rest'
LICENSE = """
Copyright © 2020  Brett Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>."""

import argparse
import enum
import logging
import operator
import os
import pkg_resources
import signal
import sys
import traceback
import types

from pathlib import Path

import requests

from typing import (
    Any,
    Callable,
    Iterator,
    NoReturn,
    Optional,
    Sequence,
    TextIO,
    Type,
    Union,
)

VERSION = pkg_resources.require(PKGNAME)[0].version

class ExceptHook:
    def __init__(self, logger: Optional[logging.Logger]=None) -> None:
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

    def __call__(self,
                 exc_type: Type[BaseException],
                 exc_value: BaseException,
                 exc_tb: types.TracebackType,
    ) -> NoReturn:
        error_type = type(exc_value).__name__
        msg = ": ".join(str(arg) for arg in exc_value.args)
        if isinstance(exc_value, KeyboardInterrupt):
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            os.kill(0, signal.SIGINT)
            signal.pause()
        elif isinstance(exc_value, requests.HTTPError):
            error_type = "PayPal API error"
            if exc_value.response is None:
                status_code = 500
            else:
                status_code = exc_value.response.status_code
            if (status_code == requests.codes.UNAUTHORIZED
                or status_code == requests.codes.FORBIDDEN):
                exitcode = os.EX_NOPERM
            elif status_code < 500:
                exitcode = os.EX_SOFTWARE
            else:
                exitcode = os.EX_UNAVAILABLE
        elif isinstance(exc_value, OSError):
            if exc_value.filename is None:
                exitcode = os.EX_OSERR
                error_type = "OS error"
                msg = exc_value.strerror
            else:
                # There are more specific exit codes for input problems vs.
                # output problems, but without knowing how the file was
                # intended to be used, we can't use them.
                exitcode = os.EX_IOERR
                error_type = "I/O error"
                msg = f"{exc_value.filename}: {exc_value.strerror}"
        else:
            exitcode = os.EX_SOFTWARE
            error_type = f"internal {error_type}"
        self.logger.critical("%s%s%s", error_type, ": " if msg else "", msg)
        self.logger.debug(
            ''.join(traceback.format_exception(exc_type, exc_value, exc_tb)),
        )
        raise SystemExit(exitcode)


class InfoAction(argparse.Action):
    def __call__(self,
                 parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: Union[Sequence[Any], str, None]=None,
                 option_string: Optional[str]=None,
    ) -> NoReturn:
        if isinstance(self.const, str):
            info = self.const
            exitcode = 0
        else:
            info, exitcode = self.const
        print(info)
        raise SystemExit(exitcode)


class LogLevel(enum.IntEnum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL
    WARN = WARNING
    ERR = ERROR
    CRIT = CRITICAL

    @classmethod
    def from_arg(cls, arg: str) -> int:
        try:
            return cls[arg.upper()].value
        except KeyError:
            raise ValueError(f"unknown loglevel {arg!r}") from None

    @classmethod
    def choices(cls) -> Iterator[str]:
        for level in sorted(cls, key=operator.attrgetter('value')):
            yield level.name.lower()


def add_loglevel_argument(parser: argparse.ArgumentParser,
                          default: LogLevel=LogLevel.INFO) -> argparse.Action:
    return parser.add_argument(
        '--loglevel',
        metavar='LEVEL',
        default=default.value,
        type=LogLevel.from_arg,
        help="Show logs at this level and above."
        f" Specify one of {', '.join(LogLevel.choices())}."
        f" Default {default.name.lower()}.",
    )

def add_version_argument(parser: argparse.ArgumentParser) -> argparse.Action:
    progname = parser.prog or sys.argv[0]
    return parser.add_argument(
        '--version', '--copyright', '--license',
        action=InfoAction,
        nargs=0,
        const=f"{progname} version {VERSION}\n{LICENSE}",
        help="Show program version and license information",
    )

def make_entry_point(mod_name: str, prog_name: str=sys.argv[0]) -> Callable[[], int]:
    """Create an entry_point function for a tool

    The returned function is suitable for use as an entry_point in setup.py.
    It sets up the root logger and excepthook, then calls the module's main
    function.
    """
    def entry_point():  # type:ignore
        prog_mod = sys.modules[mod_name]
        setup_logger()
        prog_mod.logger = logging.getLogger(prog_name)
        sys.excepthook = ExceptHook(prog_mod.logger)
        return prog_mod.main()
    return entry_point

def setup_logger(logger: Union[str, logging.Logger]='',
                 stream: TextIO=sys.stderr,
                 fmt: str='%(name)s: %(levelname)s: %(message)s',
) -> logging.Logger:
    """Set up a logger with a StreamHandler with the given format"""
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    formatter = logging.Formatter(fmt)
    handler = logging.StreamHandler(stream)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

def set_loglevel(logger: logging.Logger, loglevel: int=logging.INFO) -> None:
    """Set the loglevel for a tool or module

    If the given logger is not under a hierarchy, this function sets the
    loglevel for the root logger, along with some specific levels for libraries
    used by reporting tools. Otherwise, it's the same as
    ``logger.setLevel(loglevel)``.
    """
    if '.' not in logger.name:
        logger = logging.getLogger()
    logger.setLevel(loglevel)



#!/usr/bin/env python3
"""cliquery.py - Command line tool to query PayPal transactions and subscriptions"""
# Copyright © 2020  Brett Smith
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import collections
import datetime
import logging
import sys

from pathlib import Path
import yaml

from . import client as clientmod
from . import cliutil
from . import config as configmod
from . import paypal_types
from .transaction import CartItem, Transaction

from typing import (
    Optional,
    Sequence,
    TextIO,
    Type,
)

PROGNAME = 'paypal-query'

logger = logging.getLogger('paypal_rest.cliquery')

class YAMLDumper(yaml.SafeDumper):
    TRANSACTION_FIELD_ORDER = [
        clientmod.TransactionFields.SHIPPING,
        clientmod.TransactionFields.PAYER,
        clientmod.TransactionFields.TRANSACTION,
        clientmod.TransactionFields.CART,
        clientmod.TransactionFields.STORE,
        clientmod.TransactionFields.AUCTION,
        clientmod.TransactionFields.INCENTIVE,
    ]

    @classmethod
    def add_transaction_representer(cls, fields: clientmod.TransactionFields) -> None:
        txn_key_order = [
            f'{flag.name.lower()}_info'
            for flag in cls.TRANSACTION_FIELD_ORDER
            if flag & fields
        ]
        def transaction_representer(self: 'YAMLDumper', data: Transaction) -> yaml.nodes.MappingNode:
            return self.represent_dict((key, data[key]) for key in txn_key_order)
        cls.add_representer(Transaction, transaction_representer)


def add_fields_argument(
        parser: argparse.ArgumentParser,
        field_type: Type[clientmod.PayPalFields],
        *short_flags: str,
) -> argparse.Action:
    type_name = field_type.__name__[:-6].lower()
    return parser.add_argument(
        f'--{type_name}-fields', *short_flags,
        action='append',
        metavar='FIELD',
        type=field_type.from_arg,
        help=f"Only show these field(s) in {type_name} results."
        " You can specify this option multiple times."
        f" Choices are {', '.join(field_type.choices())}."
    )

def parse_arguments(arglist: Optional[Sequence[str]]=None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog=PROGNAME)
    cliutil.add_version_argument(parser)
    parser.add_argument(
        '--config-file', '-C',
        type=Path,
        metavar='PATH',
        help="""Read client configuration from this INI file
""")
    parser.add_argument(
        '--config-section', '-c',
        default='query',
        help="""Read client configuration from this section of the config file
""")
    parser.add_argument(
        '--begin', '--start', '-b',
        dest='start_date',
        metavar='DATETIME',
        type=paypal_types.parse_datetime,
        help="""Datetime to begin the search, in ISO 8601 format
""")
    parser.add_argument(
        '--end', '--stop', '-e',
        dest='end_date',
        metavar='DATETIME',
        type=paypal_types.parse_datetime,
        help="""Datetime to end the search, in ISO 8601 format
""")
    add_fields_argument(parser, clientmod.TransactionFields, '--txn-fields', '-T')
    add_fields_argument(parser, clientmod.SubscriptionFields, '--sub-fields', '-S')
    parser.add_argument(
        'paypal_ids',
        metavar='ID',
        nargs='*',
        help="""ID of PayPal object(s) to look up and return. If no IDs are
specified, lists all transactions in your specified date range (default last
24 hours).
""")
    cliutil.add_loglevel_argument(parser)
    args = parser.parse_args(arglist)
    args.transaction_fields = clientmod.TransactionFields.combine(args.transaction_fields)
    args.subscription_fields = clientmod.SubscriptionFields.combine(args.subscription_fields)
    if args.end_date is None:
        args.end_date = datetime.datetime.now(datetime.timezone.utc)
    return args

def summarize_transaction(txn: Transaction, stream: TextIO) -> None:
    date_s = txn.updated_date().astimezone(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M')
    status = txn.status().value
    try:
        from_s = f"\t{txn.payer_fullname()} ({txn.payer_email()})"
    except KeyError:
        from_s = ""
    print(f"{date_s}\t{txn.transaction_id()}\t{status}{from_s}", file=stream)
    cart = list(txn.cart_items())
    if not cart:
        txn_name = txn['transaction_info'].get('transaction_subject', "Gross Amount")
        txn_amt = txn.amount()
        cart.append(CartItem(None, txn_name, None, 1, txn_amt, txn_amt))
    fee_amt = txn.fee_amount()
    if fee_amt is not None:
        cart.append(CartItem(None, "PayPal Fee", None, 1, fee_amt, fee_amt))
    names = [
        item.name or item.description or item.code or "Unknown Item"
        for item in cart
    ]
    amounts = [str(item.total_price) for item in cart]
    name_len = max(len(name) for name in names)
    amt_len = max(len(amt_s) for amt_s in amounts)
    line_fmt = f'  {{:>{name_len}}} │ {{:>{amt_len}}}{{}}'
    for item, name, amt_s in zip(cart, names, amounts):
        if item.quantity != 1:
            unit_s = f" ({item.quantity:,g} @ {item.unit_price})"
        else:
            unit_s = ""
        print(line_fmt.format(name, amt_s, unit_s), file=stream)

def main(
        arglist: Optional[Sequence[str]]=None,
        stdout: TextIO=sys.stdout,
        stderr: TextIO=sys.stderr,
) -> int:
    args = parse_arguments(arglist)
    cliutil.set_loglevel(logger, args.loglevel)

    config = configmod.load_config(args.config_file)
    if args.config_section not in config:
        config.add_section(args.config_section)
    paypal = clientmod.PayPalAPIClient.from_config(config[args.config_section])

    if not args.paypal_ids:
        if args.start_date is None:
            args.start_date = args.end_date - datetime.timedelta(hours=24)
        args.transaction_fields |= clientmod.TransactionFields.TRANSACTION
        for txn in paypal.iter_transactions(
                args.start_date, args.end_date, args.transaction_fields,
        ):
            summarize_transaction(txn, stdout)
    else:
        YAMLDumper.add_transaction_representer(args.transaction_fields)
        for paypal_id in args.paypal_ids:
            paypal_id = paypal_id.upper()
            paypal_obj: paypal_types.APIResponse
            if paypal_id.startswith('I-'):
                paypal_obj = paypal.get_subscription(paypal_id, fields=args.subscription_fields)
            else:
                paypal_obj = paypal.get_transaction(
                    paypal_id, args.end_date, args.start_date, args.transaction_fields,
                )
            yaml.dump([paypal_obj], stdout, Dumper=YAMLDumper)
    return 0

entry_point = cliutil.make_entry_point(__name__, PROGNAME)

if __name__ == '__main__':
    exit(entry_point())
